import json
import logging

from cimgraph.databases import ConnectionInterface
from cimgraph.models import FeederModel, DistributedArea
import cimgraph.data_profile.cimhub_2023 as cim
import topology_processor.utils as utils

_log = logging.getLogger(__name__)

class DistributedTopologyMessage():

    def __init__(self) -> None:
        self.message = {}
        self.message['DistributionArea'] = {}

    def get_context_from_feeder(self, feeder: cim.Feeder, connection: ConnectionInterface) -> None:

        cim = connection.cim
        if not isinstance(feeder, cim.Feeder):
            raise TypeError('feeder argument should be a cim.Feeder object')
        
        feeder_model = FeederModel(container=feeder, connection=connection, distributed=True)
        feeder_area_model = feeder_model.distributed_areas[0]

        try:
            distribution_area = feeder_model.get_from_triple(feeder, 'Feeder.DistributionArea')[0]
            self.message['DistributionArea'] = json.loads(distribution_area.__repr__())
        except:
            _log.warning(f'Feeder does not have an associated DistributionArea')

        try:
            substation = feeder_model.get_from_triple(feeder, 'Feeder.NormalEnergizingSubstation')[0]
            sub_msg = json.loads(substation.__repr__())
            sub_msg['NormalEnergizedFeeder'] = []
            
            
            fdr_msg = json.loads(feeder.__repr__())
            fdr_msg['FeederArea'] = self.add_feeder_area(feeder_area_model)

            sub_msg['NormalEnergizedFeeder'].append(fdr_msg)
            self.message['DistributionArea']['Substations'] = []
            self.message['DistributionArea']['Substations'].append(sub_msg)

        except:
            _log.warning(f'Feeder does not have an associated Normal Energizing Substation')

    def get_context_from_feeder_area(self, feeder_area: cim.FeederArea,
                                     connection: ConnectionInterface) -> None:
        
        cim = connection.cim
        if not isinstance(feeder_area, cim.FeederArea):
            raise TypeError('feeder argument should be a cim.FeederArea object')
        
        feeder_area_model = DistributedArea(container=feeder_area,
                                            connection=connection, distributed=True)
        substation = feeder_area_model.get_from_triple(feeder_area, 'FeederArea.Substation')[0]
        feeders = feeder_area_model.get_from_triple(feeder_area )

        #TODO: Finish this method


    def get_context_from_distribution_area(self, distribution_area: cim.DistributionArea,
                                           connection:ConnectionInterface) -> None:
        
        cim = connection.cim
        if not isinstance(distribution_area, cim.DistributionArea):
            raise TypeError('feeder argument should be a cim.DistributionArea object')
        
        distribution_area_model = DistributedArea(container=distribution_area,
                                            connection=connection, distributed=True)
        distribution_area_model.get_all_edges(cim.DistributionArea)
        distribution_area_model.get_all_edges(cim.Substation)

        self.message['DistributionArea'] = json.loads(distribution_area.__repr__())
        self.message['DistributionArea']['Substations'] = []

        
        for substation in distribution_area.Substations:
            sub_msg = json.loads(substation.__repr__())
            sub_msg['NormalEnergizedFeeder'] = []

            for feeder in substation.NormalEnergizedFeeder:
                feeder_model = FeederModel(container=feeder, connection=connection, distributed=True)
                
                fdr_msg = json.loads(feeder.__repr__())
                # try:
                feeder_area_model = feeder_model.distributed_areas[0]
                feeder_area_model.get_all_edges(cim.FeederArea)
                fdr_msg['FeederArea'] = self.add_feeder_area(feeder_area_model)
                # except:
                #     _log.warning(f'Feeder {feeder.uri()} does not have distributed areas')

                sub_msg['NormalEnergizedFeeder'].append(fdr_msg)
                del feeder_model

            new_sub = cim.Substation()
            new_sub.uuid(uri = substation.uri())
            sub_model = DistributedArea(container=new_sub,
                                            connection=connection, distributed=False)
            sub_model.get_all_edges(cim.Substation)
            sub_model.get_all_edges(cim.ConnectivityNode)
            sub_model.get_all_edges(cim.Terminal)
            sub_msg['AddressableEquipment'] = utils.identify_addressable(sub_model)
            sub_msg['UnaddressableEquipment'] = utils.identify_unaddressable(sub_model)
            sub_msg['Measurements'] = utils.identify_measurements(sub_model)
            self.message['DistributionArea']['Substations'].append(sub_msg)
            del sub_msg
            del sub_model








    def add_feeder_area(self, feeder_area_model: DistributedArea) -> dict:
        feeder_area_model.get_all_edges(cim.FeederArea)
        feeder_area = feeder_area_model.container

        msg = json.loads(feeder_area.__repr__())
        msg['BoundaryTerminals'] = utils.identify_boundaries(feeder_area_model)
        msg['AddressableEquipment'] = utils.identify_addressable(feeder_area_model)
        msg['UnaddressableEquipment'] = utils.identify_unaddressable(feeder_area_model)
        msg['Measurements'] = utils.identify_measurements(feeder_area_model)
        msg['SwitchAreas'] = []

        for switch_area_model in feeder_area_model.distributed_areas:
            sw_msg = self.add_switch_area(switch_area_model)
            msg['SwitchAreas'].append(sw_msg)

        return msg

        

    def add_switch_area(self, switch_area_model:DistributedArea) -> dict:

        switch_area_model.get_all_edges(cim.SwitchArea)
        switch_area = switch_area_model.container
        feeder_area = switch_area_model.get_from_triple(switch_area, 'SwitchArea.FeederArea')[0]

        sw_msg = json.loads(switch_area.__repr__())
        sw_msg['FeederArea'] = json.loads(feeder_area.__repr__())
        sw_msg['BoundaryTerminals'] = utils.identify_boundaries(switch_area_model)
        sw_msg['AddressableEquipment'] = utils.identify_addressable(switch_area_model)
        sw_msg['UnaddressableEquipment'] = utils.identify_unaddressable(switch_area_model)
        sw_msg['Measurements'] = utils.identify_measurements(switch_area_model)
        sw_msg['SecondaryAreas'] = []

        for secondary_area_model in switch_area_model.distributed_areas:
            sec_msg = self.add_secondary_area(secondary_area_model)
            sw_msg['SecondaryAreas'].append(sec_msg)

        return sw_msg

    def add_secondary_area(self, secondary_area_model: DistributedArea) -> dict:

        secondary_area_model.get_all_edges(cim.SecondaryArea)
        secondary_area = secondary_area_model.container
        switch_area = secondary_area_model.get_from_triple(secondary_area, 'SecondaryArea.SwitchArea')[0]

        sa_msg = json.loads(secondary_area.__repr__())
        sa_msg['SwitchArea'] = json.loads(switch_area.__repr__())
        sa_msg['BoundaryTerminals'] = utils.identify_boundaries(secondary_area_model)
        sa_msg['AddressableEquipment'] = utils.identify_addressable(secondary_area_model)
        sa_msg['UnaddressableEquipment'] = utils.identify_unaddressable(secondary_area_model)
        sa_msg['Measurements'] = utils.identify_measurements(secondary_area_model)
        
        return sa_msg



from SPARQLWrapper import SPARQLWrapper, JSON, POST
class BlazegraphConnection(ConnectionInterface):
    sparql_obj: Optional[SPARQLWrapper] = None

    def connect(self):
        if not self.sparql_obj:
            url = self.connection_params.parameters[0].value
            self.sparql_obj = SPARQLWrapper(url)
            self.sparql_obj.setReturnFormat(JSON)

    def disconnect(self):
        self.sparql_obj = None
        
    def execute(self, query_message: str) -> QueryResponse:
        self.connect()
        self.sparql_obj.setQuery(query_message)
        self.sparql_obj.setMethod(POST)
        query_output = self.sparql_obj.query().convert()
        return query_output


import time, json
from topology_processor.topologydictionary import TopologyDictionary
from topology_processor.networkmodel import NetworkModel

class DistributedTopology():
    
    def __init__(self, gapps, model_mrid):
        self.gapps = gapps
        self.Topology = TopologyDictionary(gapps, model_mrid)
        self.MVTopology = TopologyDictionary(gapps, model_mrid)
        self.log = self.gapps.get_logger()
        
    def create_switch_areas(self, model_mrid):
        
        network = NetworkModel(self.gapps)
        self.log.info('Querying for power system model for' + str(model_mrid))
        network.build_equip_dicts(model_mrid, self.Topology)
        network.build_equip_dicts(model_mrid, self.MVTopology)

        self.log.info('Building linked lists of all equipment for ' + str(model_mrid))
        EqTypes = ['ACLineSegment', 'PowerTransformer', 'TransformerTank', 'SynchronousMachine']
        self.Topology.build_linknet(EqTypes)
        self.log.info('Building linked lists of medium-voltage equipment for ' + str(model_mrid))
        EqTypes = ['ACLineSegment', 'RatioTapChanger', 'SynchronousMachine']
        self.MVTopology.build_linknet(EqTypes)
        self.log.info('Processing switch-delimited areas for ' + str(model_mrid))
        MVTree = {}
        BreakerKeys = list(self.Topology.EquipDict['Breaker'].keys())
        MVTree = self.MVTopology.spanning_tree('Breaker', BreakerKeys , MVTree, 'all')
        FuseKeys = list(self.Topology.EquipDict['Fuse'].keys())
        MVTree = self.MVTopology.spanning_tree('Fuse', FuseKeys , MVTree, 'all')
        SwitchKeys = list(self.Topology.EquipDict['LoadBreakSwitch'].keys())
        MVTree = self.MVTopology.spanning_tree('LoadBreakSwitch', SwitchKeys, MVTree,'all')
        RecloserKeys = list(self.Topology.EquipDict['Recloser'].keys())
        MVTree = self.MVTopology.spanning_tree('Recloser', RecloserKeys, MVTree, 'all')
        
        output_message = self.create_output_message(self.Topology, MVTree, model_mrid)
        message = json.dumps(output_message)
        return message
    
    def create_output_message(self,Topology, MVTree, model_mrid):
        ConnNodeDict = Topology.ConnNodeDict
        # Initialize output message structure
        self.DistAppStruct = {}
        self.DistAppStruct['feeders'] = {}
        self.DistAppStruct['feeders']['feeder_id'] = model_mrid
        self.DistAppStruct['feeders']['addressable_equipment'] = []
        self.DistAppStruct['feeders']['unaddressable_equipment'] = []
        self.DistAppStruct['feeders']['connectivity_node'] = []
        self.DistAppStruct['feeders']['switch_areas'] = []
        ProcessedNodes = [] # List to keep track of which nodes have been processed
        SwitchKeys = list(MVTree.keys()) # Get list of all switching devices from all CIM classes
        # Iterate through all switches
        for i1 in range(len(SwitchKeys)):
            # Initialize switch area dictionary
            SwitchArea = {}
            SwitchArea['boundary_switches'] = []
            SwitchArea['addressable_equipment'] = []
            SwitchArea['unaddressable_equipment'] = []
            SwitchArea['connectivity_node'] = []
            SwitchArea['secondary_areas'] = []
            # Initialize secondary area dictionary
            DistArea = {}
            DistArea['distribution_transformer'] = []
            DistArea['addressable_equipment'] = []
            DistArea['unaddressable_equipment'] = []
            DistArea['connectivity_node'] = []
            DistAreaFlag1 = True
            for i2 in range(len(MVTree[SwitchKeys[i1]])):
                # Select next medium-voltage node, append to processed list
                node = MVTree[SwitchKeys[i1]][i2]
                ProcessedNodes.append(node)
                # Add all connected equipment
                SwitchArea['boundary_switches'].extend(ConnNodeDict[node]['Breaker'])
                SwitchArea['boundary_switches'].extend(ConnNodeDict[node]['Fuse'])
                SwitchArea['boundary_switches'].extend(ConnNodeDict[node]['LoadBreakSwitch'])
                SwitchArea['boundary_switches'].extend(ConnNodeDict[node]['Recloser'])
                SwitchArea['addressable_equipment'].extend(ConnNodeDict[node]['SynchronousMachine'])
                SwitchArea['addressable_equipment'].extend(ConnNodeDict[node]['PowerElectronicsConnection'])
                SwitchArea['addressable_equipment'].extend(ConnNodeDict[node]['LinearShuntCompensator'])
                SwitchArea['addressable_equipment'].extend(ConnNodeDict[node]['RatioTapChanger'])
                SwitchArea['addressable_equipment'].extend(ConnNodeDict[node]['EnergyConsumer'])
                SwitchArea['addressable_equipment'].extend(ConnNodeDict[node]['House'])
                SwitchArea['addressable_equipment'].extend(ConnNodeDict[node]['BatteryUnit'])
                SwitchArea['addressable_equipment'].extend(ConnNodeDict[node]['PhotovoltaicUnit'])
                SwitchArea['unaddressable_equipment'].extend(ConnNodeDict[node]['ACLineSegment'])
                SwitchArea['unaddressable_equipment'].extend(ConnNodeDict[node]['PowerTransformer'])
                SwitchArea['unaddressable_equipment'].extend(ConnNodeDict[node]['TransformerTank'])
                SwitchArea['unaddressable_equipment'].extend(ConnNodeDict[node]['Measurement'])
                SwitchArea['connectivity_node'].append(node)
                # Identify PowerTransformer and TransformerTanks for secondary areas
                DistXfmrTanks = ConnNodeDict[node]['TransformerTank'] 
                DistXfmrs = ConnNodeDict[node]['PowerTransformer']
                if DistXfmrs: # Check all PowerTransformers connected to this node
                    DistAreaFlag1 = False
                    SwitchArea['unaddressable_equipment'].extend(DistXfmrs)
                    [SwitchArea, LVNodes] = self.create_dist_area(Topology, MVTree, DistXfmrs, 'PowerTransformer', SwitchArea.copy())
                    ProcessedNodes.extend(LVNodes)
                if DistXfmrTanks: # Check all TransformerTanks connected to this node
                    DistAreaFlag1 = False
                    SwitchArea['unaddressable_equipment'].extend(DistXfmrTanks)
                    [SwitchArea, LVNodes] = self.create_dist_area(Topology, MVTree, DistXfmrTanks, 'TransformerTank', SwitchArea.copy())
                    ProcessedNodes.extend(LVNodes)


            if SwitchArea['boundary_switches']: # Append switch area if not duplicate
                self.DistAppStruct['feeders']['switch_areas'].append(dict(SwitchArea))
                self.DistAppStruct['feeders']['connectivity_node'].extend(SwitchArea['connectivity_node'])
                self.DistAppStruct['feeders']['addressable_equipment'].extend(SwitchArea['boundary_switches'])
                self.DistAppStruct['feeders']['unaddressable_equipment'].extend(SwitchArea['unaddressable_equipment'])

        # Add missing nodes to feeder level (not in switch area or secondary area)
        AllNodes = list(ConnNodeDict.keys())
        MissingNodes = list(set(AllNodes).difference(ProcessedNodes))
        for i5 in range(len(MissingNodes)):
            node = MissingNodes[i5]
            self.DistAppStruct['feeders']['addressable_equipment'].extend(ConnNodeDict[node]['SynchronousMachine'])
            self.DistAppStruct['feeders']['addressable_equipment'].extend(ConnNodeDict[node]['PowerElectronicsConnection'])
            self.DistAppStruct['feeders']['addressable_equipment'].extend(ConnNodeDict[node]['LinearShuntCompensator'])
            self.DistAppStruct['feeders']['addressable_equipment'].extend(ConnNodeDict[node]['RatioTapChanger'])
            self.DistAppStruct['feeders']['addressable_equipment'].extend(ConnNodeDict[node]['EnergyConsumer'])
            self.DistAppStruct['feeders']['addressable_equipment'].extend(ConnNodeDict[node]['House'])
            self.DistAppStruct['feeders']['addressable_equipment'].extend(ConnNodeDict[node]['BatteryUnit'])
            self.DistAppStruct['feeders']['addressable_equipment'].extend(ConnNodeDict[node]['PhotovoltaicUnit'])
            self.DistAppStruct['feeders']['unaddressable_equipment'].extend(ConnNodeDict[node]['ACLineSegment'])
            self.DistAppStruct['feeders']['unaddressable_equipment'].extend(ConnNodeDict[node]['PowerTransformer'])
            self.DistAppStruct['feeders']['unaddressable_equipment'].extend(ConnNodeDict[node]['TransformerTank'])
            self.DistAppStruct['feeders']['unaddressable_equipment'].extend(ConnNodeDict[node]['Measurement'])
            self.DistAppStruct['feeders']['connectivity_node'].append(node)

        return self.DistAppStruct
    
    def create_dist_area(self, Topology, MVTree, Xfmrs, eqtype, SwitchArea):
        # Initialize secondary area dictionary
        ConnNodeDict = Topology.ConnNodeDict
        DistArea = {}
        DistArea['distribution_transformer'] = []
        DistArea['addressable_equipment'] = []
        DistArea['unaddressable_equipment'] = []
        DistArea['connectivity_node'] = []
        LVNodes = []
        DistAreaFlag2 = False
        # Iterate through all secondary transformers
        for i3 in range(len(Xfmrs)):
            xfmr = Xfmrs[i3]
            LVTree = Topology.spanning_tree(eqtype, [xfmr], MVTree, 'all')
            LVTreeKeys = LVTree[xfmr]
            if LVTreeKeys: 
                LVTreeKeys.pop(0) # dump first node (xfmr hi-side, duplicate)
                for i4 in range(len(LVTreeKeys)):
                    lvnode = LVTreeKeys[i4]
                    DistAreaFlag2 = True
                    LVNodes.append(lvnode)
                    DistArea['distribution_transformer'] = [xfmr]
                    DistArea['addressable_equipment'].extend(ConnNodeDict[lvnode]['SynchronousMachine'])
                    DistArea['addressable_equipment'].extend(ConnNodeDict[lvnode]['PowerElectronicsConnection'])
                    DistArea['addressable_equipment'].extend(ConnNodeDict[lvnode]['EnergyConsumer'])
                    DistArea['addressable_equipment'].extend(ConnNodeDict[lvnode]['House'])
                    DistArea['addressable_equipment'].extend(ConnNodeDict[lvnode]['BatteryUnit'])
                    DistArea['addressable_equipment'].extend(ConnNodeDict[lvnode]['PhotovoltaicUnit'])
                    DistArea['unaddressable_equipment'].extend(ConnNodeDict[lvnode]['ACLineSegment'])
                    DistArea['unaddressable_equipment'].extend(ConnNodeDict[lvnode]['Measurement'])
                    DistArea['connectivity_node'].append(lvnode)
                    SwitchArea['connectivity_node'].append(lvnode)
                    SwitchArea['unaddressable_equipment'].extend(ConnNodeDict[lvnode]['ACLineSegment'])
                    SwitchArea['unaddressable_equipment'].extend(ConnNodeDict[lvnode]['Measurement'])
        if DistAreaFlag2: # append secondary area if not empty
            SwitchArea['secondary_areas'].append((DistArea.copy()))
        return SwitchArea, LVNodes


@profile
def _main():
    import os, json, time
    from gridappsd import GridAPPSD, topics as t
    from gridappsd.topics import service_input_topic, service_output_topic
    from topology_processor.distributedtopology import DistributedTopology
    from topology_processor.topologydictionary import TopologyDictionary
    from topology_processor.networkmodel import NetworkModel

    os.environ['GRIDAPPSD_APPLICATION_ID'] = 'gridappsd-topology-processor'
    os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STARTED'
    os.environ['GRIDAPPSD_USER'] = 'app_user'
    os.environ['GRIDAPPSD_PASSWORD'] = '1234App'

    # Connect to GridAPPS-D Platform
    gapps = GridAPPSD()
    assert gapps.connected


    model_mrid = "_EE71F6C9-56F0-4167-A14E-7F4C71F10EAA"

    Topology = TopologyDictionary(gapps, model_mrid)

    network = NetworkModel(gapps)

    network.build_equip_dicts(model_mrid, Topology)

    EqTypes = ['ACLineSegment', 'PowerTransformer', 'TransformerTank', 'SynchronousMachine', 'PowerElectronicsConnection']
    Topology.build_linknet(EqTypes)

    Topology.update_switches()

    Topology.build_feeder_islands()
    
            
if __name__ == "__main__":
    _main()


import json
from cimgraph.models import DistributedArea
from uuid import UUID
# import cimgraph.data_profile.cimhub_2023 as cim

jsonld = dict["@id":str(UUID),"@type":str(type)]

def identify_addressable(network: DistributedArea, 
                         equipment_list: list[jsonld] = None) -> list[jsonld]:
    
    if equipment_list is None:
        equipment_list = []

    cim = network.cim
    addressable_classes = [cim.Switch, cim.Breaker, cim.Disconnector,
        cim.Recloser, cim.LoadBreakSwitch, cim.Sectionaliser,
        cim.PowerElectronicsConnection, cim.PowerElectronicsUnit,
        cim.PhotovoltaicUnit, cim.BatteryUnit, cim.PowerElectronicsWindUnit,
        cim.RatioTapChanger, # Not technically equipment, but included for GridAPPSD
        cim.ShuntCompensator, cim.LinearShuntCompensator,
        cim.AsynchronousMachine, cim.SynchronousMachine]
    
    for class_type in addressable_classes:
        for equipment in network.graph.get(class_type, {}).values():
            # obj_json = {"@id":equipment.uri(), "@type":equipment.__class__.__name__}
            equipment_list.append(json.loads(equipment.__repr__()))
    return equipment_list

def identify_unaddressable(network: DistributedArea, 
                         equipment_list: list[jsonld] = None) -> list[jsonld]:

    if equipment_list is None:
        equipment_list = []

    cim = network.cim
    unaddressable_classes = [cim.ACLineSegment, cim.Fuse,
        cim.PowerTransformer, cim.TransformerTank,
        cim.EnergySource, cim.EnergyConsumer]
    
    for class_type in unaddressable_classes:
        for equipment in network.graph.get(class_type, {}).values():
            # obj_json = {"@id":equipment.uri(), "@type":equipment.__class__.__name__}
            equipment_list.append(json.loads(equipment.__repr__()))
    return equipment_list

def identify_measurements(network: DistributedArea, 
                         meas_list: list[jsonld] = None) -> list[jsonld]:
    
    if meas_list is None:
        meas_list = []

    cim = network.cim    
    meas_classes = [cim.Measurement, cim.Analog, cim.Discrete]
    
    for class_type in meas_classes:
        for measurement in network.graph.get(class_type, {}).values():
            meas_list.append(json.loads(measurement.__repr__()))
    return meas_list

def identify_boundaries(network: DistributedArea) -> list[jsonld]:
    
    terminals = []
    for terminal in network.container.BoundaryTerminals:
        terminals.append(json.loads(terminal.__repr__()))

    return terminals




# -------------------------------------------------------------------------------
# Copyright (c) 2022, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------
""" 
GridAPPSD Toolbox Topology Processor Library
The :mod:`gridappsd` contains a connection class :class:`gridappsd.GridAPPSD` for connecting with the
main GridAPPSD executable.
"""

from . distributedtopology import DistributedTopology
from . networkmodel import NetworkModel
from . topologydictionary import TopologyDictionary


# @dataclass
# class Parameter:
#     key: Any
#     value: Any


# @dataclass
# class ConnectionParameters:
#     parameters: List[Parameter] = field(default_factory=list)


# @dataclass
# class QueryResponse:
#     response: Any


# @dataclass
# class ConnectionInterface:
#     connection_params: ConnectionParameters

#     def connect(self):
#         raise RuntimeError("Must have implemented connect in inherited class")

#     def disconnect(self):
#         raise RuntimeError("Must have implemented disconnect in inherited class")

#     def load_attributes(self, obj: object):
#         raise RuntimeError("Must have implemented load_attributes in inherited class")

#     def create_default_instances(self, feeder_mrid: str | Feeder, mrid_list: List[str]):
#         raise RuntimeError("Must have implemented retrieve_instance from inherited class")

#     def execute(self, query: str) -> QueryResponse:
#         raise RuntimeError("Must have implemented query in the inherited class")





import time, json

class TopologyDictionary():
    
    def __init__(self, gapps, model_mrid):
        self.model_mrid = model_mrid
        self.gapps = gapps
        self.EquipDict = {}
        self.ConnNodeDict = {}
        self.TerminalsDict = {}
        self.NodeList = []
        self.TermList = []
        self.Feeders = {}
        self.Islands = {}
        self.log = self.gapps.get_logger()
        Tree = {}
        
    # Builds LinkNet linked lists for all CIM classes specified by EqTypes
    def build_linknet(self, EqTypes):
        # Intialize counter objects
        index = 0
        counter = 0
        # Build LinkNetList for all specified CIM classes:
        for i0 in range(len(EqTypes)): 
            [index, counter] = self.build_class_lists(EqTypes[i0], index, counter)
        # Add floating nodes not connected to a branch:
        StartTime = time.process_time()
        AllNodes = list(self.ConnNodeDict.keys())
        MissingNodes = list(set(AllNodes).difference(self.NodeList))
        for i1 in range(len(MissingNodes)):
            node = MissingNodes[i1]
            if 'list' not in self.ConnNodeDict[node]:
                self.ConnNodeDict[node]['node'] = index+1
                self.ConnNodeDict[node]['list'] = 0
                index = index+1
                self.NodeList.append(node)
        self.log.info("Processed " + str(len(MissingNodes)) + " missing nodes in " + str(round(1000*(time.process_time() - StartTime))) + " ms")
        # Dump JSON copies of base LinkNet structure. These are used to rebuild topo after each switch change
        self.BaseConnDict = json.dumps(self.ConnNodeDict)
        self.BaseTermDict = json.dumps(self.TerminalsDict)


    # Build LinkNet structure for single CIM equipment type, called by build_linknet()
    # Three-winding transformers not yet supported
    def build_class_lists(self, eqtype, index, old_counter):
        i2 = -1
        index2 = 0
        StartTime = time.process_time()
        EquipKeys = list(self.EquipDict[eqtype])

        for i2 in range(len(EquipKeys)):
            # Identify nodes and terminals for readability
            term1=self.EquipDict[eqtype][EquipKeys[i2]]['term1']
            node1=self.EquipDict[eqtype][EquipKeys[i2]]['node1']
            # Create keys for new terminals
            self.TerminalsDict[term1] = {}
            self.TerminalsDict[term1]['ConnectivityNode'] = node1
            self.TermList.append(term1)
            # If node1 not in LinkNet , create new keys
            if 'node' not in self.ConnNodeDict[node1]:
                self.ConnNodeDict[node1]['node'] = index+1
                self.ConnNodeDict[node1]['list'] = 0
                index = index+1
                self.NodeList.append(node1)

            # If two-terminal device, process both terminals
            if 'node2' in self.EquipDict[eqtype][EquipKeys[i2]]:
                # Identify nodes and terminals for readability
                term2=self.EquipDict[eqtype][EquipKeys[i2]]['term2']
                node2=self.EquipDict[eqtype][EquipKeys[i2]]['node2']
                # Create keys for new terminals
                self.TerminalsDict[term2] = {}
                self.TerminalsDict[term2]['ConnectivityNode'] = node2
                self.TerminalsDict[term1]['term'] = 2*i2+old_counter+1
                self.TerminalsDict[term2]['term'] = 2*i2+old_counter+2
                #self.TerminalsDict[term1]['term'] = 2*(i2+old_counter)+1
                #self.TerminalsDict[term2]['term'] = 2*(i2+old_counter)+2
                self.TermList.append(term2)
                # If node2 not in LinkNet , create new keys
                if 'node' not in self.ConnNodeDict[node2]: 
                    self.ConnNodeDict[node2]['node'] = index+1
                    self.ConnNodeDict[node2]['list'] = 0
                    index = index+1
                    self.NodeList.append(node2)
                # 1. Move node list variables to terinal next    
                self.TerminalsDict[term1]['next'] = self.ConnNodeDict[node1]['list']
                self.TerminalsDict[term2]['next'] = self.ConnNodeDict[node2]['list']
                # 2. Populate Terminal list far field with nodes
                self.TerminalsDict[term1]['far'] = self.ConnNodeDict[node2]['node']
                self.TerminalsDict[term2]['far'] = self.ConnNodeDict[node1]['node']
                # 3. Populate Connectivity nodes list with terminals
                self.ConnNodeDict[node1]['list'] = self.TerminalsDict[term1]['term']
                self.ConnNodeDict[node2]['list'] = self.TerminalsDict[term2]['term']
                index2 = index2 + 2
            # If one-terminal device, process only single terminal
            else:
                self.TerminalsDict[term1]['term'] = i2+(old_counter)+1
                #self.TerminalsDict[term1]['next'] = 0
                self.TerminalsDict[term1]['next'] = self.ConnNodeDict[node1]['list']
                #if self.ConnNodeDict[node1]['node'] == index:
                #    self.TerminalsDict[term1]['far'] = index
                #    self.ConnNodeDict[node1]['list'] = self.TerminalsDict[term1]['term']
                #else:
                self.TerminalsDict[term1]['far'] = self.ConnNodeDict[node1]['node']
                self.ConnNodeDict[node1]['list'] = self.TerminalsDict[term1]['term']
                index2 = index2 + 1

        self.log.info("Processed " + str(i2+1) + ' ' + str(eqtype) + " objects in " + str(round(1000*(time.process_time() - StartTime))) + " ms")
        counter = old_counter+index2
        return index, counter
    

    # Method to update Linknet Lists with current switch positions
    def update_switches(self):

        SwitchKeys = list(self.EquipDict['Breaker'].keys()) + list(self.EquipDict['Fuse'].keys()) + list(self.EquipDict['LoadBreakSwitch'].keys()) + list(self.EquipDict['Recloser'].keys())
        SwitchDict = {}
        SwitchDict.update(self.EquipDict['Breaker'])
        SwitchDict.update(self.EquipDict['Fuse'])
        SwitchDict.update(self.EquipDict['LoadBreakSwitch'])
        SwitchDict.update(self.EquipDict['Recloser'])

        self.ConnNodeDict = json.loads(self.BaseConnDict)
        self.TerminalsDict = json.loads(self.BaseTermDict)

        StartTime = time.process_time()
        i3 = -1
        for i3 in range(len(SwitchKeys)):

            node1=SwitchDict[SwitchKeys[i3]]['node1']
            node2=SwitchDict[SwitchKeys[i3]]['node2']


            # If switch closed, merge nodes
            if SwitchDict[SwitchKeys[i3]]['open'] == 1:
                # Merge topology Nodes
                #ConnNodeDict[node1]['TopologicalNode'] = tpnode1
                self.ConnNodeDict[node2]['TopologicalNode'] = self.ConnNodeDict[node1]['TopologicalNode'] #tpnode1
                #TopoNodeDict[tpnode1] = [node1, node2] # not implemented
                #TopoNodeDict[tpnode2] = [node2, node1]

                # Update Linked Lists
                if self.ConnNodeDict[node2]['list'] > self.ConnNodeDict[node1]['list']:
                    term2 = self.TermList[self.ConnNodeDict[node2]['list']-1]
                    next2 = self.TerminalsDict[term2]['next']
                    while next2 != 0:
                        term2 = self.TermList[next2-1]
                        next2 = self.TerminalsDict[term2]['next']
                    self.TerminalsDict[term2]['next'] = self.ConnNodeDict[node1]['list']
                    self.ConnNodeDict[node1]['list'] = self.ConnNodeDict[node2]['list']
                else:
                    term1 = self.TermList[self.ConnNodeDict[node1]['list']-1]
                    next1 = self.TerminalsDict[term1]['next']
                    while next1 != 0:
                        term1 = self.TermList[next1-1]
                        next1 = self.TerminalsDict[term1]['next']
                    self.TerminalsDict[term1]['next'] = self.ConnNodeDict[node2]['list']
                    self.ConnNodeDict[node2]['list'] = self.ConnNodeDict[node1]['list']

        self.log.info("Processed " + str(i3+1) + "switch objects in " + str(round(1000*(time.process_time() - StartTime))) + " ms")

    def build_feeder_islands(self):
        self.Feeders = {}
        self.Islands = {}
        FeederTree = {}
        fdr = -1
        isl = -1
        
        # Iterate through all PowerTransfomer objects
        XfmrKeys = list(self.EquipDict['PowerTransformer'].keys())
        StartTime = time.process_time()
        for i4 in range(len(XfmrKeys)):
            SubXfmr = XfmrKeys[i4]
            # Identify substation transfomers with high-side >34kV and low-side >1kV
            volt1 = self.EquipDict['PowerTransformer'][SubXfmr]['volt1']
            volt2 = self.EquipDict['PowerTransformer'][SubXfmr]['volt2']
            if volt1 >= 34000 and 34000 >= volt2 >= 1000:
                fdr = fdr + 1
                self.spanning_tree('PowerTransformer', [SubXfmr], FeederTree, 'single') 
                 # Add nodes to Feeder dictionary
                self.Feeders['feeder_' + str(fdr)] = {}
                self.Feeders['feeder_' + str(fdr)]['root_node'] = self.EquipDict['PowerTransformer'][SubXfmr]['node2']
                self.Feeders['feeder_' + str(fdr)]['PowerTransformer'] = SubXfmr
                self.Feeders['feeder_' + str(fdr)]['ConnectivityNode'] = FeederTree[SubXfmr] 
                # Add feeder to Node dictionary
                for i5 in range(len(FeederTree[SubXfmr])): 
                    self.ConnNodeDict[FeederTree[SubXfmr][i5]]['Feeder'].append(('feeder_' + str(fdr)))
        self.log.info('Processed ' + str(fdr + 1) + ' feeders in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")
        
        # If no suitable PowerTransformer found, build from EnergySource
        if not self.Feeders:
            Sources = list(self.EquipDict['EnergySource'].keys())
            StartTime = time.process_time()
            for i4 in range(len(Sources)):
                Sub = Sources[i4]
                fdr = fdr + 1
                self.spanning_tree('EnergySource', [Sub], FeederTree, 'single') 
                 # Add nodes to Feeder dictionary
                self.Feeders['feeder_' + str(fdr)] = {}
                self.Feeders['feeder_' + str(fdr)]['root_node'] = self.EquipDict['EnergySource'][Sub]['node1']
                self.Feeders['feeder_' + str(fdr)]['EnergySource'] = Sub
                
                self.Feeders['feeder_' + str(fdr)]['ConnectivityNode'] = FeederTree[Sub] 
                # Add feeder to Node dictionary
                for i5 in range(len(FeederTree[Sub])): 
                    self.ConnNodeDict[FeederTree[Sub][i5]]['Feeder'].append(('feeder_' + str(fdr)))
            self.log.info('Processed ' + str(fdr + 1) + ' feeders in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")
        
        # Iterate through all SynchronousMachine objects
        StartTime = time.process_time()
        IslandTree = {}
        DGKeys = list(self.EquipDict['SynchronousMachine'].keys())
        for i6 in range(len(DGKeys)):
            DG = DGKeys[i6]
            DGNode = self.EquipDict['SynchronousMachine'][DG]['node1']
            [not_in_feeder, found] = self.check_tree(DGNode, FeederTree, 'all', DG)
            if not_in_feeder:
                #IslandTree[DG] = {}
                [not_in_island, found] = self.check_tree(DGNode, IslandTree, 'all', DG)
                if not_in_island:
                    isl = isl + 1
                    self.spanning_tree('SynchronousMachine', [DG], IslandTree, 'single')
                    self.Islands['island_' + str(isl)] = {}
                    self.Islands['island_' + str(isl)]['SynchronousMachine'] = [DG]
                    self.Islands['island_' + str(isl)]['ConnectivityNode'] = IslandTree[DG] 
                    
                    # Add island to Node dictionary
                    for i7 in range(len(IslandTree[DG])): 
                        self.ConnNodeDict[IslandTree[DG][i7]]['Island'].append(('island_' + str(isl)))
                else:
                    self.Islands['island_' + str(isl)]['SynchronousMachine'].append(DGKeys[i6])
        self.log.info('Processed ' + str(isl + 1) + ' islands in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")


        # Iterate through all PowerElectronicsConnection objects
        StartTime = time.process_time()
        DERKeys = list(self.EquipDict['PowerElectronicsConnection'].keys())
        for i6 in range(len(DERKeys)):
            DER = DERKeys[i6]
            DERNode = self.EquipDict['PowerElectronicsConnection'][DER]['node1']
            [not_in_feeder, found] = self.check_tree(DERNode, FeederTree, 'all', DER)
            if not_in_feeder:
                #IslandTree[DG] = {}
                [not_in_island, found] = self.check_tree(DERNode, IslandTree, 'all', DER)
                if not_in_island:
                    isl = isl + 1
                    self.spanning_tree('PowerElectronicsConnection', [DER], IslandTree, 'single')
                    self.Islands['island_' + str(isl)] = {}
                    self.Islands['island_' + str(isl)]['PowerElectronicsConnection'] = [DER]
                    self.Islands['island_' + str(isl)]['ConnectivityNode'] = IslandTree[DER] 
                    
                    # Add island to Node dictionary
                    for i7 in range(len(IslandTree[DER])): 
                        self.ConnNodeDict[IslandTree[DER][i7]]['Island'].append(('island_' + str(isl)))
                else:
                    if 'PowerElectronicsConnection' not in self.Islands['island_' + str(isl)]:
                        self.Islands['island_' + str(isl)]['PowerElectronicsConnection'] = []
                    self.Islands['island_' + str(isl)]['PowerElectronicsConnection'].append(DERKeys[i6])
        self.log.info('Processed ' + str(isl + 1) + ' islands in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")        
    
    def spanning_tree(self, eqtype, RootKeys, Tree, Scope):
        root = ''
        TotalNodes=0
        old_len = len(Tree.keys())
        StartTime = time.process_time()

        
        for i6 in range(len(RootKeys)):
            root = RootKeys[i6]
            Tree[root] = []

            # If switch object, only use second node
            if eqtype in ['Breaker', 'Fuse', 'LoadBreakSwitch', 'Recloser']:
                [not_in_tree, found] = self.check_tree(self.EquipDict[eqtype][root]['node2'], Tree, Scope, root)
                if not_in_tree:
                    Tree[root].append(self.EquipDict[eqtype][root]['node2'])
                    FirstNode = 0
                    LastNode = 1 # only 1 node used, so initialize list at 0,1
            # If DER object, only has one node
            elif eqtype in ['SynchronousMachine', 'PowerElectronicsConnection', 'EnergySource']:
                #[not_in_tree, found] = self.check_tree(self.EquipDict[eqtype][root]['node1'], Tree, Scope, root)
               # if not_in_tree:
                Tree[root].append(self.EquipDict[eqtype][root]['node1'])
                FirstNode = 0 
                LastNode = 1 # only 1 node exists, so initialize list at 0,1
                #else:
                    
            # Otherwise, use both nodes    
            else: # Then 2-terminal object

                [not_in_tree, found] = self.check_tree(self.EquipDict[eqtype][root]['node2'], Tree, Scope, root)
                if not_in_tree:
                    Tree[root].append(self.EquipDict[eqtype][root]['node1'])
                    Tree[root].append(self.EquipDict[eqtype][root]['node2'])
                    FirstNode = 1 
                    LastNode = 2 # 2 nodes in starting list, so initialize at 1,2
                else:
                    break
            while LastNode != FirstNode:
                NextTerm = self.ConnNodeDict[Tree[root][FirstNode]]['list']
                FirstNode = FirstNode + 1
                while NextTerm != 0:
                    # Get next node and terminal for current node
                    NextNode = self.TerminalsDict[self.TermList[NextTerm-1]]['far']
                    NextTerm = self.TerminalsDict[self.TermList[NextTerm-1]]['next']
                    node = self.NodeList[NextNode-1]
                    [not_in_tree, found] = self.check_tree(node, Tree, Scope, root)
                    # Add node if not in another tree        
                    if not_in_tree:       
                        if self.ConnNodeDict[node]['nominalVoltage']:
                            # Stop building tree into sub-transmission network
                            if int(self.ConnNodeDict[node]['nominalVoltage']) < 34000: 
                                Tree[root].append(self.NodeList[NextNode-1])
                                LastNode = LastNode + 1                       
                        else: # Add node to tree if no nominal voltage defined
                            Tree[root].append(self.NodeList[NextNode-1])
                            LastNode = LastNode + 1


            self.log.info("Processed topology from  " + str(root) + ' with ' + str(len(Tree[root])) + " buses")

        if root: self.log.info("Processed " + str(len(Tree.keys()) - old_len) + " topology trees containing " + str(TotalNodes+len(Tree[root])) + " buses in " + str(round(1000*(time.process_time() - StartTime))) + " ms")

        return Tree
    
    # function to check if a node is the spanning tree
    # use argument "all" to check all trees from all root nodes
    # use argument "single" to only check the single tree from current root node
    # node is ConnectivityNode mRID to be checked
    # root is used to specify "single" tree root key 
    def check_tree(self, node, Tree, Scope, root):
        not_in_tree = True
        found = 'False'
        if Scope == 'all': 
            TreeKeys = list(Tree.keys())
            for i7 in range(len(TreeKeys)):
                if node in Tree[TreeKeys[i7]]:
                    not_in_tree = False
                    found = TreeKeys[i7]
                    break
        else: 
            if node in Tree[root]: 
                not_in_tree = False
                found = root
        return not_in_tree, found



# Custom SPARQL queries to obtain all equipment needed to build feeder topology

    
    # Get all measurements points for all equipment from Blazegraph Database
def get_all_measurements(gapps, model_mrid):
    QueryMeasurementMessage="""
        # list all measurements, with buses and equipments - DistMeasurement
        PREFIX r: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX c: <http://iec.ch/TC57/CIM100#>
        SELECT ?class ?cnid ?type ?name ?bus ?phases ?meastype ?eqname ?eqid ?trmid ?measid WHERE {
        VALUES ?fdrid {"%s"}
         ?eq c:Equipment.EquipmentContainer ?fdr.
         ?fdr c:IdentifiedObject.mRID ?fdrid. 
        { ?s r:type c:Discrete. bind ("Discrete" as ?class)}
          UNION
        { ?s r:type c:Analog. bind ("Analog" as ?class)}
         ?s c:IdentifiedObject.name ?name .
         ?s c:IdentifiedObject.mRID ?measid .
         ?s c:Measurement.PowerSystemResource ?eq .
         ?s c:Measurement.Terminal ?trm .
         ?s c:Measurement.measurementType ?type .
         ?trm c:IdentifiedObject.mRID ?trmid.
         ?eq c:IdentifiedObject.mRID ?eqid.
         ?eq c:IdentifiedObject.name ?eqname.
         #?eq r:type ?typeraw.
         # bind(strafter(str(?typeraw),"#") as ?eqtype)
         bind(strbefore(str(?name),"_") as ?meastype)
         ?trm c:Terminal.ConnectivityNode ?cn.
         ?cn c:IdentifiedObject.name ?bus.
         ?cn c:IdentifiedObject.mRID ?cnid.
         ?s c:Measurement.phases ?phsraw .
           {bind(strafter(str(?phsraw),"PhaseCode.") as ?phases)}

        } ORDER BY ?cnid ?type
        """%model_mrid

    results = gapps.query_data(query = QueryMeasurementMessage, timeout = 60)
    MeasurementQuery = results['data']['results']['bindings']
    return MeasurementQuery



# Get all ConnectivityNode and TopologicalNode objects
def get_all_nodes(gapps, model_mrid):
    QueryNodeMessage="""
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX c:  <http://iec.ch/TC57/CIM100#>
        SELECT DISTINCT ?busname ?cnid ?tpnid (group_concat(distinct ?nomu;separator="") as ?nomv ) WHERE {
        SELECT ?busname ?cnid ?tpnid ?nomu WHERE {
        VALUES ?fdrid {"%s"}
        ?fdr c:IdentifiedObject.mRID ?fdrid.
        ?bus c:ConnectivityNode.ConnectivityNodeContainer ?fdr.
        ?bus c:ConnectivityNode.TopologicalNode ?tp.
        ?bus r:type c:ConnectivityNode.
        ?bus c:IdentifiedObject.name ?busname.
        ?bus c:IdentifiedObject.mRID ?cnid.
        ?fdr c:IdentifiedObject.name ?feeder.
        ?trm c:Terminal.ConnectivityNode ?bus.
        ?trm c:Terminal.ConductingEquipment ?ce.

        OPTIONAL {
        ?ce  c:ConductingEquipment.BaseVoltage ?bv.
        ?bv  c:BaseVoltage.nominalVoltage ?nomu.
          }
        bind(strafter(str(?tp), str("http://localhost:8889/bigdata/namespace/kb/sparql#")) as ?tpnid)
        } ORDER by ?busname
        } 
        GROUP by ?busname ?cnid ?tpnid 
        ORDER by ?busname
        """%model_mrid

    results = gapps.query_data(query = QueryNodeMessage, timeout = 60)
    NodeQuery = results['data']['results']['bindings']
    return NodeQuery

# Get all switches with nodes, terminals, and default positions
def get_all_switches(gapps, model_mrid):
    QuerySwitchMessage="""
        # list nodes for Breakers, Reclosers, LoadBreakSwitches, Fuses, Sectionalisers in a selected feeder
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX c:  <http://iec.ch/TC57/CIM100#>
        SELECT ?cimtype ?name ?id ?bus1 ?bus2 ?term1 ?term2 ?node1 ?node2 ?tpnode1 ?tpnode2 ?open (group_concat(distinct ?phs;separator="") as ?phases) WHERE {
        SELECT ?cimtype ?name ?id ?bus1 ?bus2 ?term1 ?term2 ?node1 ?node2 ?tpnode1 ?tpnode2 ?phs ?open WHERE {
        VALUES ?fdrid {"%s"}  # 13 bus
        VALUES ?cimraw {c:LoadBreakSwitch c:Recloser c:Breaker c:Fuse c:Sectionaliser}
        ?fdr c:IdentifiedObject.mRID ?fdrid.
        ?s r:type ?cimraw.
        bind(strafter(str(?cimraw),"#") as ?cimtype)
        ?s c:Equipment.EquipmentContainer ?fdr.
        ?s c:IdentifiedObject.name ?name.
        ?s c:IdentifiedObject.mRID ?id.
        ?s c:Switch.normalOpen ?open.
        ?t1 c:Terminal.ConductingEquipment ?s.
        ?t1 c:ACDCTerminal.sequenceNumber "1".
        ?t1 c:Terminal.ConnectivityNode ?cn1. 
        ?cn1 c:ConnectivityNode.TopologicalNode ?tp1.
        ?cn1 c:IdentifiedObject.name ?bus1.
        ?t2 c:Terminal.ConductingEquipment ?s.
        ?t2 c:ACDCTerminal.sequenceNumber "2".
        ?t2 c:Terminal.ConnectivityNode ?cn2. 
        ?cn2 c:ConnectivityNode.TopologicalNode ?tp2.
        ?cn2 c:IdentifiedObject.name ?bus2
            OPTIONAL {?swp c:SwitchPhase.Switch ?s.
            ?swp c:SwitchPhase.phaseSide1 ?phsraw.
            bind(strafter(str(?phsraw),"SinglePhaseKind.") as ?phs) }
            bind(strafter(str(?t1), "#") as ?term1) 
            bind(strafter(str(?t2), "#") as ?term2)
            bind(strafter(str(?cn1), "#") as ?node1)
            bind(strafter(str(?cn2), "#") as ?node2)
            bind(strafter(str(?tp1), "#") as ?tpnode1)
            bind(strafter(str(?tp2), "#") as ?tpnode2)
        } ORDER BY ?name ?phs
        }
        GROUP BY ?cimtype ?name ?id ?bus1 ?bus2 ?term1 ?term2 ?node1 ?node2 ?tpnode1 ?tpnode2 ?open
        ORDER BY ?cimtype ?name
        """%model_mrid
    results = gapps.query_data(query = QuerySwitchMessage, timeout = 60)
    SwitchQuery = results['data']['results']['bindings']
    return SwitchQuery

def get_all_transformers(gapps,model_mrid):
    QueryXfmrMessage="""
        # list all the terminals connected to a TransformerEnd for CIMWriter
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX c:  <http://iec.ch/TC57/CIM100#>
        SELECT ?class ?eqid ?endid ?tname ?tid ?bus ?cnid ?tpid ?seq ?phs ?ratedu WHERE {
        VALUES ?fdrid {"%s"} 
        ?fdr c:IdentifiedObject.mRID ?fdrid.
        {?pxf c:Equipment.EquipmentContainer ?fdr.
        ?end c:PowerTransformerEnd.PowerTransformer ?pxf.
        ?end c:PowerTransformerEnd.ratedU ?ratedu.
        ?pxf c:IdentifiedObject.mRID ?eqid.
        }
        UNION
        {?tank c:Equipment.EquipmentContainer ?fdr.
        ?end c:TransformerTankEnd.TransformerTank ?tank.
        ?tank c:IdentifiedObject.mRID ?eqid.
        ?end c:TransformerTankEnd.phases ?ph.
        }
        ?end c:TransformerEnd.Terminal ?t.
        ?t c:Terminal.ConnectivityNode ?cn. 
        ?t c:IdentifiedObject.name ?tname.

        ?cn c:ConnectivityNode.TopologicalNode ?tp.
        ?cn c:IdentifiedObject.name ?bus.
        ?t c:ACDCTerminal.sequenceNumber ?seq.
        bind(strafter(str(?end),"#") as ?endid).
        bind(strafter(str(?t),"#") as ?tid).
        bind(strafter(str(?cn),"#") as ?cnid).
        bind(strafter(str(?tp),"#") as ?tpid).
        bind(strbefore(str(?endclass), "E") as ?class)
        bind(strafter(str(?ph),"e.") as ?phs).
        ?end a ?classraw.
        bind(strafter(str(?classraw),"CIM100#") as ?endclass)
        }
        ORDER by ?class ?eqid ?tname ?endid ?bus ?cnid ?tpid ?seq ?phs ?ratedu
        """%model_mrid
    results = gapps.query_data(query = QueryXfmrMessage, timeout = 60)
    XfmrQuery = results['data']['results']['bindings']
    return XfmrQuery

def get_all_lines(gapps, model_mrid):
    QueryLineMessage="""
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX c:  <http://iec.ch/TC57/CIM100#>
        SELECT ?name ?bus1 ?bus2 ?id ?tname1 ?term1 ?tname2 ?term2 ?node1 ?node2 ?tpnode1 ?tpnode2 (group_concat(distinct ?phs;separator="") as ?phases) WHERE {
        SELECT ?name ?bus1 ?bus2 ?phs ?id ?tname1 ?term1 ?tname2 ?term2 ?node1 ?node2 ?tpnode1 ?tpnode2 WHERE {
        VALUES ?fdrid {"%s"}  
        ?fdr c:IdentifiedObject.mRID ?fdrid.
        ?s r:type c:ACLineSegment.
        ?s c:Equipment.EquipmentContainer ?fdr.
        ?s c:IdentifiedObject.name ?name.
        ?s c:IdentifiedObject.mRID ?id.
        ?t1 c:Terminal.ConductingEquipment ?s.
        ?t1 c:ACDCTerminal.sequenceNumber "1".
        ?t1 c:Terminal.ConnectivityNode ?cn1. 
        ?t1 c:IdentifiedObject.name ?tname1.
        ?cn1 c:IdentifiedObject.name ?bus1.
        ?cn1 c:ConnectivityNode.TopologicalNode ?tp1.
        ?t2 c:Terminal.ConductingEquipment ?s.
        ?t2 c:ACDCTerminal.sequenceNumber "2".
        ?t2 c:Terminal.ConnectivityNode ?cn2. 
        ?t2 c:IdentifiedObject.name ?tname2.
        ?cn2 c:ConnectivityNode.TopologicalNode ?tp2.
        ?cn2 c:IdentifiedObject.name ?bus2.
        bind(strafter(str(?t),"#") as ?tid).
            bind(strafter(str(?t1), "#") as ?term1) 
            bind(strafter(str(?t2), "#") as ?term2)
            bind(strafter(str(?cn1), "#") as ?node1)
            bind(strafter(str(?cn2), "#") as ?node2)
            bind(strafter(str(?tp1), "#") as ?tpnode1)
            bind(strafter(str(?tp2), "#") as ?tpnode2)
                OPTIONAL {?acp c:ACLineSegmentPhase.ACLineSegment ?s.
                ?acp c:ACLineSegmentPhase.phase ?phsraw.
                bind(strafter(str(?phsraw),"SinglePhaseKind.") as ?phs) }

        } ORDER BY ?name ?phs
        }
        GROUP BY ?name ?bus1 ?bus2 ?id ?tname1 ?term1 ?tname2 ?term2 ?node1 ?node2 ?tpnode1 ?tpnode2
        ORDER BY ?name
        """%model_mrid
    results = gapps.query_data(query = QueryLineMessage, timeout = 60)
    LineQuery = results['data']['results']['bindings']
    return LineQuery

def get_all_houses(gapps, model_mrid):
    QueryHouseMessage = """
    # list houses - DistHouse
    PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX c:  <http://iec.ch/TC57/CIM100#>
    SELECT  ?name ?parent ?id ?cnid ?tid
    WHERE { 
    VALUES ?fdrid {"%s"} 
       ?h r:type c:House.
       ?h c:IdentifiedObject.name ?name.
       ?h c:IdentifiedObject.mRID ?id.
       ?h c:House.EnergyConsumer ?econ.
       ?econ c:IdentifiedObject.name ?parent.
       ?t c:Terminal.ConductingEquipment ?econ.
       ?t c:Terminal.ConnectivityNode ?cn.
       ?fdr c:IdentifiedObject.mRID ?fdrid.
       ?fdr c:IdentifiedObject.name ?fdrname.
       ?econ c:Equipment.EquipmentContainer ?fdr.
       bind(strafter(str(?t),"#") as ?tid).
       bind(strafter(str(?cn),"#") as ?cnid).
    } ORDER BY ?name
    """%model_mrid
    results = gapps.query_data(query = QueryHouseMessage, timeout = 60)
    HouseQuery = results['data']['results']['bindings']
    return HouseQuery

def get_all_tapchangers(gapps, model_mrid):
    QueryTapMessage = """
    # voltage regulators - DistRegulator
    PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX c:  <http://iec.ch/TC57/CIM100#>
    SELECT ?rname ?eqid ?wnum ?cnid ?tid ?seq ?pxfid ?tankid
    WHERE {
    VALUES ?fdrid {"%s"}  # 123 PV
     ?pxf c:Equipment.EquipmentContainer ?fdr.
     ?fdr c:IdentifiedObject.mRID ?fdrid.
     ?rtc r:type c:RatioTapChanger.
     ?rtc c:IdentifiedObject.name ?rname.
     ?rtc c:RatioTapChanger.TransformerEnd ?end.
     ?end c:TransformerEnd.endNumber ?wnum.
     ?end c:TransformerEnd.Terminal ?t.
     ?t c:Terminal.ConnectivityNode ?cn. 
     ?t c:ACDCTerminal.sequenceNumber ?seq.
     {?end c:PowerTransformerEnd.PowerTransformer ?pxf.
      ?pxf c:IdentifiedObject.mRID ?eqid.}
     UNION
     {?end c:TransformerTankEnd.TransformerTank ?tank.
     ?tank c:IdentifiedObject.mRID ?eqid.
     ?tank c:TransformerTank.PowerTransformer ?pxf.}
     bind(strafter(str(?t),"#") as ?tid).
     bind(strafter(str(?cn),"#") as ?cnid).
     bind(strafter(str(?pxf),"#") as ?pxfid).
     bind(strafter(str(?tank),"#") as ?tankid).
    }
    ORDER BY ?rname ?wnum
    """%model_mrid
    results = gapps.query_data(query = QueryTapMessage, timeout = 60)
    TapChangerQuery = results['data']['results']['bindings']
    return TapChangerQuery

def get_all_energy_sources(gapps, model_mrid):
    QuerySourceMessage = """
    # substation source - DistSubstation
    PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX c:  <http://iec.ch/TC57/CIM100#>
    SELECT ?name ?bus ?basev ?nomv ?node ?term ?source WHERE {
     ?s r:type c:EnergySource.
    # feeder selection options - if all commented out, query matches all feeders
    VALUES ?fdrid {"%s"}
     ?s c:Equipment.EquipmentContainer ?fdr.
     ?fdr c:IdentifiedObject.mRID ?fdrid.
     ?s c:IdentifiedObject.name ?name.
     ?s c:ConductingEquipment.BaseVoltage ?bv.
     ?bv c:BaseVoltage.nominalVoltage ?basev.
     ?s c:EnergySource.nominalVoltage ?nomv.  
     ?t c:Terminal.ConductingEquipment ?s.
     ?t c:Terminal.ConnectivityNode ?cn. 
     ?cn c:IdentifiedObject.name ?bus
     bind(strafter(str(?t), "#") as ?term)
     bind(strafter(str(?cn), "#") as ?node)
     bind(strafter(str(?s), "#") as ?source)
    }
    ORDER by ?name
    """%model_mrid
    results = gapps.query_data(query = QuerySourceMessage, timeout = 60)
    SourceQuery = results['data']['results']['bindings']
    return SourceQuery

def get_all_batteries(gapps, model_mrid):
    QueryBattMessage = """
    # Storage - DistStorage
    PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX c:  <http://iec.ch/TC57/CIM100#>
    SELECT ?name ?eqid ?fdrid ?pecid ?termid ?cnid WHERE {
     ?s r:type c:BatteryUnit.
     ?s c:IdentifiedObject.name ?name.
     ?pec c:PowerElectronicsConnection.PowerElectronicsUnit ?s.

     VALUES ?fdrid {"%s"}  
     ?pec c:Equipment.EquipmentContainer ?fdr.
     ?fdr c:IdentifiedObject.mRID ?fdrid.
     ?pec c:IdentifiedObject.mRID ?pecid.
     bind(strafter(str(?s),"#") as ?eqid).
     ?t c:Terminal.ConductingEquipment ?pec.
     ?t c:Terminal.ConnectivityNode ?cn. 
     bind(strafter(str(?cn),"#") as ?cnid).
     ?t c:IdentifiedObject.mRID ?termid.

    }
    GROUP by ?name ?eqid ?fdrid ?pecid ?termid ?cnid
    ORDER by ?name
    """%model_mrid
    results = gapps.query_data(query = QueryBattMessage, timeout = 60)
    BattQuery = results['data']['results']['bindings']
    return BattQuery

def get_all_photovoltaics(gapps, model_mrid):
    QueryPVMessage = """
    PREFIX r: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX c: <http://iec.ch/TC57/CIM100#>
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?name ?eqid ?pecid ?cnid ?termid
    WHERE {
    VALUES ?feeder_mrid {"%s"}
    ?s r:type c:PhotovoltaicUnit.
    ?s c:IdentifiedObject.name ?name.
    ?s c:IdentifiedObject.mRID ?eqid.
    ?pec c:PowerElectronicsConnection.PowerElectronicsUnit ?s.
    ?pec c:IdentifiedObject.mRID ?pecid.
    ?pec c:Equipment.EquipmentContainer ?fdr.
    ?fdr c:IdentifiedObject.mRID ?feeder_mrid.
    ?t c:Terminal.ConductingEquipment ?pec.
    ?t c:Terminal.ConnectivityNode ?cn.
    bind(strafter(str(?cn),"#") as ?cnid).
    ?t c:IdentifiedObject.mRID ?termid.
    }
    GROUP by ?name ?eqid ?pecid ?cnid ?termid
    ORDER by ?name
    """%model_mrid
    results = gapps.query_data(query = QueryPVMessage, timeout = 60)
    PVQuery = results['data']['results']['bindings']
    return PVQuery


from gridappsd import GridAPPSD
import time, json
import topology_processor.topo_meas_queries as queries

class NetworkModel(GridAPPSD):

    def __init__(self, gapps):
        self.gapps = gapps
        self.log = self.gapps.get_logger()
    
    # This method builds equipment dictionaries for given model and Topology object
    def build_equip_dicts(self, model_mrid, Topology):
        # Initialize all equipment dictionary keys
        Topology.EquipDict['ACLineSegment'] = {}
        Topology.EquipDict['BatteryUnit'] = {}
        Topology.EquipDict['Breaker'] = {}
        Topology.EquipDict['EnergyConsumer'] = {}
        Topology.EquipDict['House'] = {}
        Topology.EquipDict['Fuse'] = {}
        Topology.EquipDict['LinearShuntCompensator'] = {}
        Topology.EquipDict['LoadBreakSwitch'] = {}
        Topology.EquipDict['PhotovoltaicUnit'] = {}
        Topology.EquipDict['PowerTransformer'] = {}
        Topology.EquipDict['RatioTapChanger'] = {}
        Topology.EquipDict['Recloser'] = {}
        Topology.EquipDict['TransformerTank'] = {}    
        Topology.EquipDict['SynchronousMachine'] = {}
        Topology.EquipDict['PowerElectronicsConnection'] = {}
        Topology.EquipDict['EnergySource'] = {}

        # Initialize dictionary keys for all ConnectivityNode objects in model:
        StartTime = time.process_time()
        i0=-1
        NodeQuery=queries.get_all_nodes(self.gapps,model_mrid)
        for i0 in range(len(NodeQuery)):
            node=NodeQuery[i0]['cnid']['value']
            Topology.ConnNodeDict[node] = {}
            Topology.ConnNodeDict[node]['name'] = NodeQuery[i0]['busname']['value']
            Topology.ConnNodeDict[node]['TopologicalNode'] = NodeQuery[i0]['tpnid']['value']
            if 'nomv' in NodeQuery[i0]:
                Topology.ConnNodeDict[node]['nominalVoltage'] = NodeQuery[i0]['nomv']['value']
            else:
                Topology.ConnNodeDict[node]['nominalVoltage'] = []
            Topology.ConnNodeDict[node]['ACLineSegment'] = []
            Topology.ConnNodeDict[node]['BatteryUnit'] = []
            Topology.ConnNodeDict[node]['Breaker'] = []
            Topology.ConnNodeDict[node]['EnergyConsumer'] = []
            Topology.ConnNodeDict[node]['Fuse'] = []
            Topology.ConnNodeDict[node]['House'] = []
            Topology.ConnNodeDict[node]['LinearShuntCompensator'] = []
            Topology.ConnNodeDict[node]['LoadBreakSwitch'] = []
            Topology.ConnNodeDict[node]['PhotovoltaicUnit'] = []
            Topology.ConnNodeDict[node]['PowerTransformer'] = []
            Topology.ConnNodeDict[node]['RatioTapChanger'] = []
            Topology.ConnNodeDict[node]['Recloser'] = []
            Topology.ConnNodeDict[node]['TransformerTank'] = []    
            Topology.ConnNodeDict[node]['SynchronousMachine'] = []
            Topology.ConnNodeDict[node]['PowerElectronicsConnection'] = []
            Topology.ConnNodeDict[node]['Measurement'] = []
            Topology.ConnNodeDict[node]['Feeder'] = []
            Topology.ConnNodeDict[node]['Island'] = []
        self.log.info('Processed ' + str(i0+1) + ' ConnectivyNode objects in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")

        # Import all measurements and associated objects:
        StartTime = time.process_time()
        MeasurementQuery=queries.get_all_measurements(self.gapps,model_mrid)
        i1 = -1
        # Parse all entries in query response
        for i1 in range(len(MeasurementQuery)):    
            node = MeasurementQuery[i1]['cnid']['value']
            eqtype = MeasurementQuery[i1]['meastype']['value']
            eqid = MeasurementQuery[i1]['eqid']['value']
            meastype = MeasurementQuery[i1]['type']['value']
            measid = MeasurementQuery[i1]['measid']['value']
            # Associate measurement mRID with ConnectivityNode
            Topology.ConnNodeDict[node]['Measurement'].append(MeasurementQuery[i1]['measid']['value'])
            # Associate equipment mRID with ConnectivityNode if not already defined by prior measurement
            if eqid not in Topology.ConnNodeDict[node][eqtype]: 
                Topology.ConnNodeDict[node][eqtype].append(eqid)
            # Create equipment dictionary entry if not already defined by prior measurement
            if eqid not in Topology.EquipDict[eqtype]: 
                Topology.EquipDict[eqtype][eqid] = {}
                #Topology.EquipDict[eqtype][eqid]['meas'] = {}
                #Topology.EquipDict[eqtype][eqid]['meas']['PNV'] = []
                #Topology.EquipDict[eqtype][eqid]['meas']['VA'] = []
                #Topology.EquipDict[eqtype][eqid]['meas']['Pos'] = []
                #Topology.EquipDict[eqtype][eqid]['meas']['A'] = []
                #Topology.EquipDict[eqtype][eqid]['meas']['SoC'] = []
                
            # Associate measurement value with equipment
            #Topology.EquipDict[eqtype][eqid]['meas'][meastype].append(measid)
            
            # Associate ConnectivityNode with equipment mRID - FIRST PASS
            if 'node1' in Topology.EquipDict[eqtype][eqid]: # If one node already defined, then assume two-terminal branch
                if Topology.EquipDict[eqtype][eqid]['node1'] != node:
                    Topology.EquipDict[eqtype][eqid]['node2'] = node
                    Topology.EquipDict[eqtype][eqid]['term2'] = MeasurementQuery[i1]['trmid']['value']
            else: # If first node, assume that it is first node
                Topology.EquipDict[eqtype][eqid]['name'] = MeasurementQuery[i1]['eqname']['value']
                Topology.EquipDict[eqtype][eqid]['node1'] = node
                Topology.EquipDict[eqtype][eqid]['term1'] = MeasurementQuery[i1]['trmid']['value']
            # NEED TO ADD LOGIC FOR 3-WINDING TRANSFORMERS LATER
        self.log.info('Processed ' + str(i1+1) + ' Measurement objects in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")

        # Import all ACLineSegment objects - SECOND PASS
        StartTime = time.process_time()
        LineQuery = queries.get_all_lines(self.gapps, model_mrid)
        i2 = -1
        eqtype = 'ACLineSegment'
        for i2 in range(len(LineQuery)):
            eqid = LineQuery[i2]['id']['value']

            # Associate equipment mRID with ConnectivityNode if not already defined by prior measurement
            if eqid not in Topology.ConnNodeDict[LineQuery[i2]['node1']['value']][eqtype]: 
                Topology.ConnNodeDict[LineQuery[i2]['node1']['value']][eqtype].append(eqid)
            if eqid not in Topology.ConnNodeDict[LineQuery[i2]['node2']['value']][eqtype]: 
                Topology.ConnNodeDict[LineQuery[i2]['node2']['value']][eqtype].append(eqid)
            Topology.EquipDict[eqtype][eqid]['term1'] = LineQuery[i2]['term1']['value']
            Topology.EquipDict[eqtype][eqid]['term2'] = LineQuery[i2]['term2']['value']
            Topology.EquipDict[eqtype][eqid]['node1'] = LineQuery[i2]['node1']['value']
            Topology.EquipDict[eqtype][eqid]['node2'] = LineQuery[i2]['node2']['value']
        self.log.info('Processed ' + str(i2+1) + ' ACLineSegment objects in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")

        # Import all PowerTransformer and TransformerTank objects - SECOND PASS
        StartTime = time.process_time()
        XfmrQuery = queries.get_all_transformers(self.gapps, model_mrid)
        i2 = -1
        for i2 in range(len(XfmrQuery)):
            eqtype = XfmrQuery[i2]['class']['value']
            eqid = XfmrQuery[i2]['eqid']['value']
            seq = str(XfmrQuery[i2]['seq']['value'])
            # Check if transformer not defined when parsing measurements
            if eqid not in Topology.EquipDict[eqtype]: Topology.EquipDict[eqtype][eqid] = {}
            # Identify terminal sequence and create keys for new terminals
            Topology.EquipDict[eqtype][eqid]['bus' + seq] = XfmrQuery[i2]['bus']['value']
            Topology.EquipDict[eqtype][eqid]['term' + seq] = XfmrQuery[i2]['tid']['value']
            Topology.EquipDict[eqtype][eqid]['node' + seq] = XfmrQuery[i2]['cnid']['value']
            Topology.EquipDict[eqtype][eqid]['tname' + seq] = XfmrQuery[i2]['tname']['value']
            if eqid not in Topology.ConnNodeDict[XfmrQuery[i2]['cnid']['value']][eqtype]:
                Topology.ConnNodeDict[XfmrQuery[i2]['cnid']['value']][eqtype].append(eqid)
            if 'ratedu' in XfmrQuery[i2]: # Add rated voltage if defined
                Topology.EquipDict[eqtype][eqid]['volt' + seq] = int(float(XfmrQuery[i2]['ratedu']['value']))
            else: Topology.EquipDict[eqtype][eqid]['volt' + seq] = 0 
            if 'phs' in XfmrQuery[i2]:  # Add phase if defined
                Topology.EquipDict[eqtype][eqid]['phase' + seq] = XfmrQuery[i2]['phs']['value'] 
            else: Topology.EquipDict[eqtype][eqid]['phase' + seq] = {}
        self.log.info('Processed ' + str(i2+1) + ' Transformer objects in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")

        # Import all Breaker, Fuse, LoadBreakSwitch, and Recloser objects -  SECOND PASS
        StartTime = time.process_time()
        SwitchQuery = queries.get_all_switches(self.gapps, model_mrid)
        i3 = -1
        for i3 in range(len(SwitchQuery)):
            eqid = SwitchQuery[i3]['id']['value']
            eqtype = SwitchQuery[i3]['cimtype']['value']
            # Check if switch not defined when parsing measurements
            if eqid not in Topology.EquipDict[eqtype]: Topology.EquipDict[eqtype][eqid] = {}
            Topology.EquipDict[eqtype][eqid]['term1']=SwitchQuery[i3]['term1']['value']
            Topology.EquipDict[eqtype][eqid]['term2']=SwitchQuery[i3]['term2']['value']
            Topology.EquipDict[eqtype][eqid]['node1']=SwitchQuery[i3]['node1']['value']
            Topology.EquipDict[eqtype][eqid]['node2']=SwitchQuery[i3]['node2']['value']
            if eqid not in Topology.ConnNodeDict[SwitchQuery[i3]['node1']['value']][eqtype]:
                Topology.ConnNodeDict[SwitchQuery[i3]['node1']['value']][eqtype].append(eqid)
            if eqid not in Topology.ConnNodeDict[SwitchQuery[i3]['node2']['value']][eqtype]:
                Topology.ConnNodeDict[SwitchQuery[i3]['node2']['value']][eqtype].append(eqid)
            # Check if switch is open or closed in base model
            if SwitchQuery[i3]['open']['value'] == 'false': 
                Topology.EquipDict[eqtype][eqid]['open'] = 1
            else: 
                Topology.EquipDict[eqtype][eqid]['open'] = 0
        self.log.info('Processed ' + str(i3+1) + ' Switch objects in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")


        # Import all House objects
        StartTime = time.process_time()
        HouseQuery = queries.get_all_houses(self.gapps, model_mrid)
        i4 = -1
        for i4 in range(len(HouseQuery)):
            eqid = HouseQuery[i3]['id']['value']
            eqtype = 'House'
            # Check if house not defined when parsing measurements
            if eqid not in Topology.EquipDict[eqtype]: Topology.EquipDict[eqtype][eqid] = {}
            Topology.EquipDict[eqtype][eqid]['term1']=HouseQuery[i3]['tid']['value']
            Topology.EquipDict[eqtype][eqid]['node1']=HouseQuery[i3]['cnid']['value']

        self.log.info('Processed ' + str(i4+1) + ' House objects in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")

        # Import all RatioTapChanger objects
        StartTime = time.process_time()
        TapChangerQuery = queries.get_all_tapchangers(self.gapps, model_mrid)
        eqtype = 'RatioTapChanger'
        i5 = -1
        for i5 in range(len(TapChangerQuery)):
            eqid = TapChangerQuery[i5]['pxfid']['value']

            if 'tankid' in TapChangerQuery[i5]:
                tankid = TapChangerQuery[i5]['tankid']['value']
                if tankid in Topology.EquipDict['TransformerTank']:                     
                    Topology.EquipDict[eqtype][eqid] = dict(Topology.EquipDict['TransformerTank'][tankid])
                    Topology.EquipDict[eqtype][eqid]['TransformerTank'] = tankid
                    
            if 'pxfid' in TapChangerQuery[i5]:
                pxfid = TapChangerQuery[i5]['pxfid']['value']
                if pxfid in Topology.EquipDict['PowerTransformer']:
                    Topology.EquipDict[eqtype][eqid] = dict(Topology.EquipDict['PowerTransformer'][pxfid])
                    Topology.EquipDict[eqtype][eqid]['PowerTransformer'] = pxfid
                    
            Topology.EquipDict[eqtype][eqid]['name'] = TapChangerQuery[i5]['rname']['value']
 
        self.log.info('Processed ' + str(i5+1) + ' RatioTapChanger objects in ' + str(round(1000*(time.process_time() - StartTime))) + " ms")
        
        # Import all EnergySource objects
        startTime = time.process_time()
        SourceQuery = queries.get_all_energy_sources(self.gapps, model_mrid)
        eqtype = 'EnergySource'
        for i6 in range(len(SourceQuery)):
            eqid = SourceQuery[i6]['source']['value']
            if eqid not in Topology.EquipDict[eqtype]:
                Topology.EquipDict[eqtype][eqid] = {}
            Topology.EquipDict[eqtype][eqid]['name'] = SourceQuery[i6]['name']['value']
            Topology.EquipDict[eqtype][eqid]['node1'] = SourceQuery[i6]['node']['value']
            Topology.EquipDict[eqtype][eqid]['term1'] = SourceQuery[i6]['term']['value']
            
        # Import all BatteryUnit objects
        startTime = time.process_time()
        BattQuery = queries.get_all_batteries(self.gapps, model_mrid)
        eqtype = 'BatteryUnit'
        for i7 in range(len(BattQuery)):
            eqid = BattQuery[i7]['eqid']['value']
            node = BattQuery[i7]['cnid']['value']
            if eqid not in Topology.EquipDict[eqtype]:
                Topology.EquipDict[eqtype][eqid] = {}
            Topology.EquipDict[eqtype][eqid]['name'] = BattQuery[i7]['name']['value']
            Topology.EquipDict[eqtype][eqid]['node1'] = BattQuery[i7]['cnid']['value']
            Topology.EquipDict[eqtype][eqid]['term1'] = BattQuery[i7]['termid']['value']
            # Associate equipment mRID with ConnectivityNode if not already defined by prior measurement
            if eqid not in Topology.ConnNodeDict[node][eqtype]: 
                Topology.ConnNodeDict[node][eqtype].append(eqid)
            
        # Import all PhotovoltaicUnit objects
        startTime = time.process_time()
        PVQuery = queries.get_all_photovoltaics(self.gapps, model_mrid)
        eqtype = 'PhotovoltaicUnit'
        for i7 in range(len(PVQuery)):
            eqid = PVQuery[i7]['eqid']['value']
            node = PVQuery[i7]['cnid']['value']
            if eqid not in Topology.EquipDict[eqtype]:
                Topology.EquipDict[eqtype][eqid] = {}
            Topology.EquipDict[eqtype][eqid]['name'] = PVQuery[i7]['name']['value']
            Topology.EquipDict[eqtype][eqid]['node1'] = PVQuery[i7]['cnid']['value']
            Topology.EquipDict[eqtype][eqid]['term1'] = PVQuery[i7]['termid']['value']
            # Associate equipment mRID with ConnectivityNode if not already defined by prior measurement
            if eqid not in Topology.ConnNodeDict[node][eqtype]: 
                Topology.ConnNodeDict[node][eqtype].append(eqid)
            
